;;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Indent-tabs-mode: NIL -*-
;;;
;;; scheduler.lisp --- Controlling the queue of scheduled events
;;;                    and running expired timers.
;;;
;;; Copyright (C) 2003 Zach Beane <xach@xach.com>
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining
;;; a copy of this software and associated documentation files (the
;;; "Software"), to deal in the Software without restriction, including
;;; without limitation the rights to use, copy, modify, merge,publish,
;;; distribute, sublicense, and/or sell copies of the Software, and to
;;; permit persons to whom the Software is furnished to do so, subject to
;;; the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be
;;; included in all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
;;; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
;;; MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
;;; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
;;; LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
;;; OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
;;; WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

(in-package :bordeaux-timers)


;;;
;;; Control messages
;;;

(defun make-control-message (command timer)
  (cons command timer))

(defun control-message-command (message)
  (car message))

(defun control-message-timer (message)
  (cdr message))


;;;
;;; Control stack
;;;

(defvar *control-stack* (make-array 10
                                    :adjustable t
                                    :fill-pointer 0))

(defvar *control-stack-lock* (bt:make-lock "Control stack lock"))
(defvar *control-stack-waitqueue* (bt:make-condition-variable))

(defmacro with-control-stack-lock (&body body)
  `(bt:with-lock-held (*control-stack-lock*)
     ,@body))

(defun next-control-message ()
  (with-control-stack-lock
    (loop (if (zerop (length *control-stack*))
              (bt:condition-wait *control-stack-waitqueue*
                                 *control-stack-lock*)
              (return-from next-control-message
                (vector-pop *control-stack*))))))

(defun add-control-message (message)
  (with-control-stack-lock
    (vector-push-extend message *control-stack*))
  (bt:condition-notify *control-stack-waitqueue*))


;;;
;;; Public interface (all manipulate the control stack)
;;;

(defun schedule-timer (timer absolute-time &optional repeat-time)
  (setf (%timer-expire-time timer) (universal-time-to-precise-time absolute-time)
        (%timer-repeat-time timer) repeat-time)
  (add-control-message (make-control-message :schedule timer))
  (values))

(defun schedule-timer-relative (timer relative-time &optional repeat-time)
  (setf (%timer-expire-time timer) (+ (get-precise-time) relative-time)
        (%timer-repeat-time timer) repeat-time)
  (add-control-message (make-control-message :schedule timer)))

(defun unschedule-timer (timer)
  (add-control-message (make-control-message :unschedule timer)))


;;; Not public, but related

(defun reschedule-timer (timer)
  (setf (%timer-expire-time timer) (+ (get-precise-time)
                                      (%timer-repeat-time timer)))
  (add-control-message (make-control-message :reschedule timer)))


;;;
;;; The scheduler
;;;

(defvar *schedule* (make-priority-queue :key #'%timer-expire-time))

(defun peek-schedule ()
  (priority-queue-maximum *schedule*))

(defun process-control-message (message)
  (ecase (control-message-command message)
    ((:schedule :reschedule)
     (priority-queue-insert *schedule* (control-message-timer message)))
    (:unschedule
     (let ((timer (control-message-timer message)))
       (priority-queue-remove *schedule* timer)
       (with-control-stack-lock
           (setf *control-stack*
                 (delete-if #'(lambda (message)
                                (and (eq (control-message-command message)
                                         :reschedule)
                                     (eq (control-message-timer message)
                                         timer)))
                            *control-stack*)))))))

(defun time-to-next-timer ()
  (let ((timer (peek-schedule)))
    (when timer
      (- (%timer-expire-time timer) (get-precise-time)))))

(defmacro with-optional-timeout (delay &body body)
  (let ((d (gensym "DELAY"))
        (body-fun (gensym "BODY-FUNC")))
    `(let ((,d ,delay))
       (flet ((,body-fun () ,@body))
         (cond ((null ,d)
                (,body-fun))
               ((<= ,d (float (/ 1 internal-time-units-per-second))) ; XXX NOT 0
                (error 'bt:timeout))
               (t (bt:with-timeout (,d)
                    (,body-fun))))))))

(defun scheduler ()
  (loop
     (expire-pending-timers)
     (let ((delay (time-to-next-timer)))
       (handler-case
           (with-optional-timeout delay
             (let ((message (next-control-message)))
               (process-control-message message)))
         (bt:timeout () nil)))))


;;;
;;; Expiring timers
;;;

(defun expire-timer (timer)
  (with-slots (function repeat-time)
      timer
    (if (%timer-thread timer)
        (bt:make-thread function)
        (funcall function))
    (when repeat-time
      (reschedule-timer timer))))

(defun expire-pending-timers ()
  (loop
     (let ((next-timer (peek-schedule)))
       (unless next-timer
         (return-from expire-pending-timers))
       (if (> (get-precise-time) (%timer-expire-time next-timer))
           (expire-timer (priority-queue-extract-maximum *schedule*))
           (return-from expire-pending-timers)))))


;;;
;;; Starting the system
;;;

(defvar *timers-enabled-p* nil)
(defvar *timers-enabled-mutex* (bt:make-lock "Timer lock"))

(defun enable-timers ()
  (bt:with-lock-held (*timers-enabled-mutex*)
    (unless *timers-enabled-p*
      (setf *timers-enabled-p* t)
      (bt:make-thread #'(lambda ()
                          (unwind-protect
                               (scheduler)
                            ;; XXX this doesn't seem to run if
                            ;;  the thread is killed
                            (setf *timers-enabled-p* nil))))
      (values))))

(defun timers-enabled-p ()
  *timers-enabled-p*)
